/* Try to solve a polynomial, using home-grown techniques.
 * By: Dirk Mittler
 * Date: October 25, 2019
 *
 * First Windows-patch added: May 16, 2020
 * Further Windows-patches added until: June 2, 2020
 *
 * Refinement:
 * To deal with the instability of deflation, employ a second pass.
 *
 * The instability of deflation is such, that small, allowed
 * inaccuracies in early roots, leads to successively less accurate,
 * deflated polynomials to approximate. The fact that each of their
 * roots can be found with equally accurate or inaccurate
 * precision, does not make up for the fact that, at that point, the
 * deflated polynomial itself has become inaccurate. Thus, accurate
 * roots were being approximated, for inaccurate polynomials, at first
 * with no way to compensate.
 *
 * Proposed solution:
 * Resume computations with the original polynomial, but this time,
 * Sort the found roots into DESCENDING ORDER, and then,
 * factor out the found roots in reverse order, POLISHING EACH TIME.
 *
 * Refinement first conceived on June 5, 2020 by Dirk Mittler.
 *
 *
 * Refinement:
 * Problem:
 * If given an absurdly large constant term, i.e., (-1e36), and a
 * polynomial of the 12th degree to solve, the program may
 * output some obviously false roots, also with an order of
 * magnitude of ~10^36. The suspected reason this happens is, the
 * purported root being raised to the power of (11), which results
 * in the CPU computing infinity, and assigning this to (epsilond),
 * inside 'search_root()'. Once this corrupted result has been
 * computed, it may both pass and fail whatever set of inequality-tests
 * it needs to, to become validated as a root.
 *
 * Proposed Solution:
 * The test can be performed within 'search_root()', whether
 * (epsilond) is in fact infinite, and if so, to return 'false'.
 *
 * Refinement first conceived on August 22, 2022 by Dirk Mittler.
 * 
 * Afterthought, August 23, 2022:
 * In reality, this malfunction was not purely a failure to test,
 * whether to raise a real number to a power produces infinity.
 * It requires as much that my 'apply_poly()' function, when
 * multiplying complex root-candidates, produces a complex number
 * which is infinite. After all, just as it goes in pure Math, in C++,
 * 
 * (infinity < infinity)
 * 
 * Will produce false.
 *
 */

#ifdef _WIN32

#define WINVER 0x0601
#define _WIN32_WINNT 0x0601

#define M_PI 3.14159265358979323846264338328
/*
 * Obviously, because the number written above is only a
 * 'double' Literal, the full series of 30 decimal places
 * will not get used. The equivalent precision is to
 * 14 decimal places.
 *
 */

#endif

#include <cmath>
#include <complex>
#include <algorithm>
#include <iostream>
#include <iomanip>
#include <stdlib.h>
#include <string.h>

// Bug fix added May 14, 2022...
#include <limits>

#define MIN_A0 0.01

#ifndef max
#define max(a, b) ((a > b) ? a : b)
#endif

using std::complex;
using namespace std::complex_literals;
using std::cout;
using std::endl;
using std::setprecision;

typedef complex<long double> CC;

struct Polynomial {
    CC *coeffs;
    int deg;
    int offset;
    CC *roots;
    int solved;
    int o_deg;
};

void normalize(Polynomial *a)
{
    int array_end = a->offset + a->deg;
    CC first_coeff = a->coeffs[a->offset];
    a->coeffs[a->offset] = 1.0;
    for (int i = a->offset + 1; i <= array_end; i++) {
        a->coeffs[i] /= first_coeff;
    }

    return;
}

//  Apply polynomial to parameter.
CC apply_poly(Polynomial *a, CC param)
{

    int array_end = a->offset + a->deg;

    CC power = 1.0;
    CC result = (a->coeffs)[array_end];
    for (int i = 1; i <= a->deg; i++) {
        power = 1.0;
        for (int j = 1; j <= i; j++) {
            power *= param;
        }

        result += power * (a->coeffs)[array_end - i];
    }

    return result;
}

//  In-place division of polynomial.
//  Returns degree of remaining plynomial in same array.
int divide(Polynomial *a, CC root, CC *remaind)
{
    CC multiple = 0;
    int array_end = a->offset + a->deg;
    for (int i = a->offset; i < array_end; i++) {
        multiple = a->coeffs[i];
        a->coeffs[i + 1] += multiple * root;
    }

    *remaind = a->coeffs[array_end];
    a->deg--;

    return a->deg;
}

//  Augmented division of polynomial,
//  may or may not polish.
bool try_root(Polynomial *a, CC *root, bool polish,
              long double epsilon = 1.0E-12)
{
    CC adjustment = 0.0;
    CC remaind = 0.0;
    CC deriv = 0.0;

    if (a->deg < 1)
        return false;
    int N = a->deg;

    double deriv2_a0 = 0.0;

    int passes = 1;
    if (N == 1)
        polish = false;
    if (polish)
        passes = 3;
    int array_end = a->offset + N;

    Polynomial quot;
    quot.coeffs = new CC[array_end + 1];

    while (passes-- > 0) {
		
		using namespace std;
		
        *root += adjustment;

        quot.offset = a->offset;
        quot.deg = N;

        for (int i = 0; i <= array_end; i++) {
            quot.coeffs[i] = a->coeffs[i];
        }

        //  (1.0)^0 is possible here.
        //	!! Some libraries' power functions cast
        //	their first arguments to 'double' (from 'long double').
        deriv2_a0 = N * N * pow(norm(*root) + 1.0, N - 1);
        //	If what was computed here is infinite, verification
        //	will not take place, but, polishing still will.

        //	This will be treated as an invalid result...
		#ifdef isinf
        if (isinf(deriv2_a0)) {
            delete[] quot.coeffs;
			return false;
		}
		#endif

        if (norm(apply_poly(&quot, *root)) >
                epsilon * epsilon * deriv2_a0) {
            delete[] quot.coeffs;
            return false;
        }

        divide(&quot, *root, &remaind);

        adjustment = CC(0.0, 0.0);

        deriv = -apply_poly(&quot, *root);

        if (polish && norm(deriv) > abs(remaind)) {
            adjustment = remaind / deriv;
        }
    }

    for (int i = 0; i < array_end; i++) {
        a->coeffs[i] = quot.coeffs[i];
    }
    a->deg = quot.deg;
    a->offset = quot.offset;

    delete[] quot.coeffs;

    return true;
}

//  Solver for linear case.
//  Assumes that the equation has been normalized.
inline bool solve_lin(Polynomial *a, CC *root)
{
    if (a->deg != 1)
        return false;

    *root = -a->coeffs[a->offset + 1];
    return true;
}

//  Do the actual work.
bool search_root(Polynomial *a, CC *root, long double epsilon = 1.0E-12)
{
    int N = a->deg;

    if (N < 1)
        return false;

    //  !!
    if (N == 1)
        return solve_lin(a, root);

    CC xa = -1.0;
    //	CC xb = exp(CC(0.0, M_PI / (N + 1)));
    CC xb = exp((M_PI / (N + 1)) * 1i);

    CC ya = apply_poly(a, xa);
    CC yb = apply_poly(a, xb);

    double spread = abs(yb - ya);
    double spread2 = 0.0;

    if (N <= 4 && pow(spread, N) < epsilon) {
        return false;
    } else if (N > 4 && pow(spread, N - 2) < N * N * epsilon) {
        return false;
    }

    int misses = 0;

    long double epsilond = epsilon;
    CC t = 0.0;
    CC xt = 0.0;

    //  Really bad form: while (true)
    while (true) {
		
		using namespace std;
		
        epsilond = N * (pow(abs(xa), N - 1)) * epsilon;
        if (epsilond < epsilon / N)
            epsilond = epsilon / N;

		#ifdef isinf
        if (isinf(epsilond))
            return false;
        #endif
        
        /* At this point in program flow, 3 situations could exist...
         * 
         * Power function does not cast 'long double' to 'double':
         * 
         * 'long double' can hold the powers of root candidates
         * without becoming infinite. Arithmetic continues as though
         * Algebra.
         * 
         * sizeof(double) == sizeof(long double):
         * 
         * (epsilond * epsilond) could be infinite, but if it is,
         * norm(ya) would need to be finite to generate
         * root vicinity. In that case, a hypothetical abs(ya)
         * would also be less than (epsilond).
         * 
         * sizeof(double) < sizeof(long double):
         * 
         * Because (epsilond) holds a value which is finite according
         * to 'double', (epsilond * epsilond) will be finite according
         * to 'long double'. No problem here, either.
         * 
         */

        if (norm(ya) < epsilond * epsilond) {
            *root = xa;
            return true;
        }

        if (norm(ya - yb) <= epsilon * epsilon * epsilon) {
            return false;
        }

        t = (-ya) / (yb - ya);

        xt = (t * xb) + ((CC(1.0, 0.0) - t) * xa);

        xb = xt + (CC(0.0, 0.5) * (xa - xt));
        xa = xt;

        ya = apply_poly(a, xa);
        yb = apply_poly(a, xb);

        spread2 = abs(ya);

        if (spread2 >= spread) {
            misses++;
        }

        if ((misses > N + 2) || (spread2 != spread2)) {
            return false;
        }

        spread = spread2;
    }
}

//  Return number of derivatives...
int create_deriv_list(Polynomial a, Polynomial **l)
{

    CC coeff = 0.0;
    *l = new Polynomial[max(a.deg, 1)];

    Polynomial *L = *l;

    int g_offset = a.offset;
    int array_end = a.deg + g_offset;
    L[0].deg = a.deg;
    L[0].o_deg = a.deg;
    L[0].offset = g_offset;
    L[0].coeffs = new CC[array_end + 1];
    L[0].solved = 0;
    L[0].roots = new CC[a.deg];

    for (int i = 0; i <= array_end; i++) {
        L[0].coeffs[i] = a.coeffs[i];
    }

    if (abs(L[0].coeffs[g_offset]) < MIN_A0) {
        L[0].coeffs[g_offset] = CC(1.0, 0.0);
    }

    normalize(L);

    for (int i = 1; i < a.deg; i++) {
        L[i].deg = L[i-1].deg - 1;
        L[i].o_deg = L[i].deg;
        L[i].offset = g_offset;
        L[i].coeffs = new CC[array_end + 1];
        L[i].solved = 0;
        L[i].roots = new CC[L[i].deg];
        for (int j = 0; j < L[i-1].deg; j++) {

            coeff = L[i-1].coeffs[L[i].deg + g_offset - j];
            L[i].coeffs[L[i].deg + g_offset - j] = coeff * CC(1.0 + j, 0.0);

        }

        normalize(L + i);

    }

    return max(a.deg - 1, 0);
}

void destroy_deriv_list(Polynomial *l)
{
    int array_num = l[0].o_deg;
    for (int i = 0; i < array_num; i++) {
        delete[] l[i].coeffs;
        delete[] l[i].roots;
    }

    delete[] l;
    return;
}

bool climb_root(Polynomial *L, CC *root, int i, int mult,
                long double epsilon = 1.0E-12)
{
    int s = 0;
    int mult_temp = 0;
    bool last_attempt = false;

    if (i < 0)
        return true;

    while ((last_attempt = try_root(L + i, root, false, epsilon))) {
        s = L[i].solved;
        L[i].roots[s] = *root;
        L[i].solved++;
        if (++mult_temp >= mult)
            break;
    }

    if (!last_attempt)
        return false;

    return climb_root(L, root, i - 1, mult + 1, epsilon);
}

// Define absolute of real number.
// Should not be necessary.
template <class T>
T abs_d(T x)
{
    if (x < 0)
        return -x;
    return x;
}

//  Bubble-Sort roots of solved poly into descending order
void sort_roots(Polynomial *solved_p)
{
    CC temp(0.0, 0.0);
    int deg = solved_p->solved;
    for (int i = 0; i < deg; i++) {
        for (int j = 0; j < i; j++) {
            if (norm(solved_p->roots[j]) < norm(solved_p->roots[i])) {
                temp = solved_p->roots[j];
                solved_p->roots[j] = solved_p->roots[i];
                solved_p->roots[i] = temp;
            }
        }
    }
}

//  Factor out solved roots in reverse order
void reverse_factor(Polynomial *target_p, Polynomial *solved_p,
                    long double epsilon = 1.0E-11)
{
    int deg = solved_p->solved;
    int offset = target_p->offset;
    if (abs(target_p->coeffs[offset]) < MIN_A0)
        target_p->coeffs[offset] = CC(1.0, 0.0);

    normalize(target_p);

    sort_roots(solved_p);

    CC temp_root = solved_p->roots[--deg];

    while (deg > 0) {
        if (try_root(target_p, &temp_root, true, epsilon)) {
            solved_p->roots[deg--] = temp_root;
        } else {
            solved_p->roots[deg--] = CC(0.0, 0.0);
        }
        temp_root = solved_p->roots[deg];
    }
}

//  Finally, the driver:
bool poly_solve(Polynomial *a, Polynomial **PL, bool polish,
                bool reverse)
{
    create_deriv_list(*a, PL);
    int s = 0;
    long double epsilon = 1.0E-12;
    Polynomial *L = *PL;

    CC current_root = 0.0;
    CC current_conjugate = 0.0;

    if (sizeof(CC) == 32) {
        epsilon = 1.0L;
        int limits_10 = std::numeric_limits<long double>::digits10;
        for (int i = 3; i < limits_10; i++)
            epsilon /= 10.0L;
    }

    for (int i = a->deg - 1; i >= 0; i--) {
        while (search_root(L + i, &current_root, epsilon) &&
                try_root(L + i, &current_root, polish, epsilon)) {

            s = L[i].solved;
            L[i].roots[s] = current_root;
            L[i].solved++;

            climb_root(L, &current_root, i - 1, 2, epsilon);

            if (abs_d(imag(current_root)) > epsilon) {

                current_conjugate = conj(current_root);

                if (try_root(L + i, &current_conjugate, false, epsilon)) {

                    s = L[i].solved;
                    L[i].roots[s] = current_conjugate;
                    L[i].solved++;

                    climb_root(L, &current_conjugate, i - 1, 2, epsilon);
                }
            }
        }
    }

    //  New feature as of June 5, 2020:
    //  Factor out solutions in reverse, while
    //  polishing:

    if (reverse)
        reverse_factor(a, L, epsilon);

    //  Improvised feature:
    //  Print solutions.
    //  This wouldn't be done in actual software.
    //  I.e., Edit out.
    cout << endl;
    for (int i = 0; i < L[0].solved; i++) {
        current_root = L[0].roots[i];

        //  Secondary Polishing...
        if (abs_d(imag(current_root)) <= epsilon) {
            current_root = CC(real(current_root), 0.0);
        }
        if (abs_d(real(current_root)) <= epsilon) {
            current_root = CC(0.0, imag(current_root));
        }

        cout << real(current_root) << "  +  ";
        cout << imag(current_root) << "I";
        cout << endl;
    }

    cout << endl;
    cout << endl;

    return true;
}

int main(int argc, char *argv[])
{
    Polynomial *PL = NULL;

    Polynomial A;
    char instring[] = ".";

    CC temp = 0.0;
    char *comma = NULL;

    int arg_offset = 0;
    bool polish = true;
    bool reverse = false;

    if (argc > 1 && *argv[1] == '!') {
        arg_offset = 1;
        polish = false;
    }

    if (argc > 1 && strcmp(argv[1], "+") == 0) {
        arg_offset = 1;
        polish = true;
        reverse = true;
        cout << endl;
        cout << "Forwards Deflation was more stable, when roots were" << endl;
        cout << "found in order of increasing magnitude." << endl;
        cout << "Reverse Deflation is more stable when roots are" << endl;
        cout << "found in order of decreasing magnitude." << endl;
        cout << endl;
    }

    cout << setprecision(14);

    if (sizeof(CC) == 32) {
        cout << setprecision(std::numeric_limits<long double>::digits10 - 2);
    }

    if (argc - arg_offset < 3) {
        cout << "Usage:" << endl;
        cout << endl;
        cout << "rroots [!|+] coeff1 coeff2 coeff3 coeff4 ..." << endl;
        cout << endl;
        cout << "In the above case, prints the solutions to:" << endl;
        cout << "coeff1*x^3 + coeff2*x^2 + coeff3*x + coeff4 = 0.0" << endl;
        cout << "All command-line arguments must be numbers," << endl;
        cout << "separated by spaces. The only way to omit a term," << endl;
        cout << "is to put a zero." << endl;
        cout << "Minimum degree: 1 (a linear equation)." << endl;
        cout << endl;
        cout << "Because all the polynomials need to be normalized," << endl;
        cout << "i.e., the coefficient of the highest exponent" << endl;
        cout << "made equal to 1.0, values deemed too low as the first" << endl;
        cout << "argument, will arbitrarily be set to +1.0 ." << endl;
        cout << endl;
        cout << "Precede coefficients-list with an exclamation-mark," << endl;
        cout << "i.e., with a ! , to turn polishing off." << endl;
        cout << "Sometimes, polishing can break finding roots." << endl;
        cout << endl;
        cout << "Imaginary components can be added to coefficients, by typing" << endl;
        cout << "a comma after the real component, followed without space" << endl;
        cout << "by the imaginary component(s)." << endl;
        cout << endl;

        cout << "Instead of an exclamation-mark, precede coefficients-list" << endl;
        cout << "with a plus-sign, i.e., with a + , to factor out the list" << endl;
        cout << "of found roots in reverse order, while polishing," << endl;
        cout << "in order to test and compensate for the instability" << endl;
        cout << "normally inherent in forwards-deflation." << endl;
        cout << "(May give more-accurate results for some polynomials.)" << endl;
        cout << endl;

        cout << "Type Enter, to terminate this program." << endl;
        std::cin.get(instring, 1);

        return 1;
    }

    //  The following line may be edited out:
    cout << "RRoots_Dirk_Mittler, Copyright (C) 2020, Dirk Mittler" << endl;

    //  The following lines fulfill 'Appropriate License Notification':
    cout << "This program comes with ABSOLUTELY NO WARRANTY." << endl;
    cout << "This is free software, and you are welcome to redistribute it" << endl;
    cout << "under certain conditions." << endl;
    cout << "The license under which this program has been copyrighted, can be found" << endl;
    cout << "At:  https://www.gnu.org/licenses/gpl-3.0.txt" << endl;
    cout << endl;

    A.offset = 0;
    A.deg = argc - arg_offset - 2;

    A.coeffs = new CC[A.deg + 1];

    for (int c = 0; c <= A.deg; c++) {
        temp = CC(strtod(argv[c + arg_offset + 1], &comma), 0.0);
        if (*comma == ',') {
            temp = CC(real(temp), strtod(comma + 1, NULL));
        }

        A.coeffs[c] = temp;
    }

    poly_solve(&A, &PL, polish, reverse);

    destroy_deriv_list(PL);

    delete[] A.coeffs;

    return 0;
}
